#include <iostream>
#include <memory>
#include <vector>

using namespace std;

class Money { public: virtual ~Money() {} };
class Bread : public Money { public: virtual ~Bread() {} };
class Dinero : public Money { public: virtual ~Dinero() {} };

template< typename Money >
class Processor {
public:
  void process(Money const&) {
    cout << "processing money" << endl;
  }
};

template<>
class Processor<Bread> {
public:
  void process(Bread const&) {
    cout << "processing bread" << endl;
  }
};

template<>
class Processor<Dinero> {
public:
  void process(Dinero const&) {
    cout << "processing dinero" << endl;
  }
};

class I_processor {
public:
  virtual void process(Money const&) = 0;
};

class Bread_processor : I_processor {
public:

  inline void process(Bread const& bread) {
    process(static_cast<Money const&>(bread));
  }

  inline void process(Money const& bread) {
    cout << "virtually processing bread by calling ct" << endl;
    impl_.process(dynamic_cast<Bread const&>(bread));
  }
  Processor<Bread> impl_;
};

class Dinero_processor : I_processor {
public:
  inline void process(Dinero const& dinero) {
    process(static_cast<Money const&>(dinero));
  }

  void process(Money const& dinero) {
    cout << "virtually processing dinero by calling ct" << endl;
    impl_.process(dynamic_cast<Dinero const&>(dinero));
  }
  Processor<Dinero> impl_;
};


int main(int argc, char **argv) {

  {
    Processor<Bread> processor;
    processor.process(Bread{});
  }

  {
    Processor<Dinero> processor;
    processor.process(Dinero{});
  }

  {
    Bread_processor processor;
    processor.process(Bread{});
  }

  {
    Dinero_processor processor;
    processor.process(Dinero{});
  }

  using Money_ptr = shared_ptr<Money>;
  using Money_clip = vector<Money_ptr>;

  Money_clip money_clip;
  money_clip.push_back(Money_ptr(new Bread()));
  money_clip.push_back(Money_ptr(new Dinero()));

  std::cout << "done" << std::endl;
  return 0;
}
